import fiona
import geopandas as gpd

fiona.drvsupport.supported_drivers['kml'] = 'rw' # enable KML support which is disabled by default
fiona.drvsupport.supported_drivers['KML'] = 'rw' # enable KML support which is disabled by default

# with open(kml_file) as f:

#      doc = kmlparser.parse(f)
    
def get_kml_chunk_bounds(kml_file, chunks):
    """
    Slow function to read that big old KML file on the sentinel site and strip out boundaries. 

    KML file is in afloat-extras\data\sentinel\spatial_chunk_kml or can be found on their website. 

    Locations of interest:
        UWA 2023 SWOT moorings:
            - Middle of 51LVE
            - Western edge of 51LWE
        Scott reef:
            - 51LUE
        Broome:
            - T51KVA

    """
    print('Reading KML file.')
    my_map = gpd.read_file(kml_file, driver='KML')
    my_map
    print('....done.')

    my_latbounds = {}
    my_lonbounds = {}
    
    chunk_col = my_map.Name.values
    
    for chunk in chunks:
        ci = chunk_col == chunk
        my_chunk = my_map[ci]

        my_gemometry = my_chunk['geometry'].values[0]
        type(my_gemometry)
        dir(my_gemometry)
        my_bounds = my_gemometry.bounds
        print(my_bounds)

        my_lonbounds[chunk] = [my_bounds[0], my_bounds[2], my_bounds[2], my_bounds[0], my_bounds[0]]
        my_latbounds[chunk] = [my_bounds[1], my_bounds[1], my_bounds[3], my_bounds[3], my_bounds[1]]

    return my_lonbounds, my_latbounds


"""
Various tools for quality controling or otherwise preprocessing turbulenec timeseries data, including:

Functions
=========
    - unwrap:               phase unwrapping pulse coherent velocity records
    - goring_nikora_2002:   despike by the method of goring and nikora 2002

For more conventional QAQC use the pIMOS library. 

These functions work on continuous individual blocks of data. If they are to be done blockwise 
this must be handled elsewhere.   

"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import path as mplpath

def unwrap(x, prev_block_median, ambiguity_velocity=0.23, verbose = False):
    """
    Unwrap a block of data, or don't if not needed:

    Inputs
    ======

        - x: The block of data
        - prev_block_median: the median of the previous block, to help determine whether its wrapped high or low. 

    Outputs
    =======

        - thats_a_wrap: boolean, true if wrapped, false if not wrapped
        - X: output data. equal to  x if the block was not deemed to be wrapped, or an unwrapped version otherwise. 

    """

    def detect_wrap(x, prev_block_median, ambiguity_velocity, verbose = False):
        """
        Detect whether wrap is suspected. Method is to do a test unwrap (direction does not matter) 
        and compare the new variance with the old. If it goes down it's assumed to be wrapped. 
        """
        
        vb = np.nanvar(x)

        X = unwrap_block(x, prev_block_median, ambiguity_velocity, verbose = verbose)

        thats_a_wrap = np.nanvar(X) < np.nanvar(x) # Check based on variance shift
        thats_a_wrap = np.max(np.diff(X[~np.isnan(X)])) < max(np.diff(x[~np.isnan(x)]))
        
        if verbose:
            if thats_a_wrap:
                print('Var X = {}; var x = {}; Wrapped'.format(np.nanvar(X), vb))
            else:
                print('Var X = {}; var x = {}; Not wrapped wooo!!'.format(np.nanvar(X), vb))

        if verbose and thats_a_wrap:
            plt.figure(figsize=(12, 8))

            ax = plt.subplot2grid((3, 20), (0, 0), rowspan=1, colspan=29)
            ax.plot(x, linewidth=3)
            ax.plot(X)
            ax.grid(linestyle='-')
            plt.show()

        return thats_a_wrap, X

    def unwrap_block(x, prev_block_median, ambiguity_velocity, verbose = False):
        """
        Only having auto detection from now on. 
        """
        
        X  = x.copy()
        X_ = X.copy()

        sx = np.sort(x)
        db = np.diff(sx);                           # Measure "spikes"
        mdb = np.max(db);                           # Take largest "spike"
        dbi = int(np.mean(np.nonzero(db==mdb)))     # Get index of largest "spike"

        ind = np.arange(dbi, dbi+2)

        brk = np.mean(sx[ind])                      # Find largest "spike" mid velocity
        
        median_above = np.nanmedian(x[x > brk])
        median_below = np.nanmedian(x[x < brk])

        if verbose:
            print('Largest spike is {} m/s, from {} to {}, midpoint is {}'.format(mdb, sx[ind[0]], sx[ind[1]], brk))
            print('Median of all points above(below) the break is {}({})'.format(median_above, median_below))
        
        if np.abs(median_above-prev_block_median) < np.abs(median_below-prev_block_median): # Modified this line fro orig
            hlt = 'low';
            direction = 'down';
        else:
            hlt = 'high';
            direction = 'up';
        
        # print('It is thus assumed that the {} values are phase wrapped.'.format(hlt))

        if direction == 'down':
                phase_wrapped = x < brk
                
                X[phase_wrapped] = x[phase_wrapped] + 2*ambiguity_velocity;
        if direction == 'up':
                phase_wrapped = x > brk
                
                X[phase_wrapped] = x[phase_wrapped] - 2*ambiguity_velocity;
        
        if verbose:
            print('n changed = {} (of {}).'.format(sum(phase_wrapped), len(phase_wrapped)))

        return X

    # Main function
    x = x.copy()

    thats_a_wrap, X = detect_wrap(x, prev_block_median, ambiguity_velocity=ambiguity_velocity, verbose = verbose)
    if thats_a_wrap:
        pass
    else:
        X = x 
        
    if verbose:
        if thats_a_wrap:
            print('I''m not wrapped but it is wrapped')
        else:
            print('I''m wrapped that it isn''t wrapped')

    return thats_a_wrap, X 
 
def despike(u, fs, interp_opt='linear', intensity = 0.9, verbose=False):
    """
    Despike a block of data. Binning to be done outside. Same as despikei but returns the first output only.  
    """

    U, di = despikei(u, fs, interp_opt='linear', intensity = 0.9, verbose=False)

    return U

def despikei(u, fs, interp_opt='linear', intensity = 0.9, verbose=False):
    """
    Despike a block of data. Binning to be done outside.  

    """

    def phase_space_thres(u, fs, intensity = 0.9, verbose=False):
        """
        This function actually detects the spikes and nans them out. Single block, single component. 
        """

         # not oriignally part of the method. 

        n = len(u)
        u = u - u.mean(0) # Maybe should detrend

        du = np.zeros_like(u)
        d2u = np.zeros_like(u)
        
        # Take the centered difference.
        du[1:-1] = (u[2:] - u[:-2]) / 2
        # And again.
        d2u[2:-2] = (du[1:-1][2:] - du[1:-1][:-2]) / 2
        
        std_u = np.std(u, axis=0)
        std_du = np.std(du, axis=0)
        std_d2u = np.std(d2u, axis=0)
        
        # goringnikora(data,tplot)
        ## Part 1 of Phase-Threshold Method
        # Eq 7 Nikora & Goring

        ## Part 2 of Phase Threshold Method

        # Expected absolute maximum using Eq.2
        Lambda_u = (2*np.log(n))**0.5
        em_u    =   Lambda_u * std_u; 
        em_du   =   Lambda_u * std_du; 
        em_d2u  =   Lambda_u * std_d2u; 

        ## Part 3 of Phase Threshold Method
        # Eq.9
        theta = np.arctan(np.nansum(u*std_d2u)/np.nansum(u**2))

        if verbose:
            print('Theta = {}'.format(theta))

        A = np.array([[np.cos(theta)**2, np.sin(theta)**2], [np.sin(theta)**2, np.cos(theta)**2]])
        LHS = np.array([[em_u**2], [em_d2u**2]])
        ab2 = np.linalg.inv(A) @ LHS
        a = np.sqrt(ab2[0])
        b = np.sqrt(ab2[1])

        def index_from_ellipse(x, y, r_x, r_y, phi):

            origin_x = np.mean(x)
            origin_y = np.mean(y)

            xp, yp  = calculate_ellipse(origin_x, origin_y, r_x, r_y, phi)

            xy_1 = np.transpose(np.vstack((xp, yp)))
            p = mplpath.Path(xy_1)
            points = np.transpose(np.vstack((x, y)))
            ind = ~p.contains_points(points)

            return xp, yp, ind

        # d2u vs u
        r_x = intensity*a
        r_y = intensity*b
        xp_1, yp_1, ind_1 = index_from_ellipse(u, d2u, r_x, r_y, theta)

        # du vs u
        r_x = intensity*em_u
        r_y = intensity*em_du
        xp_2, yp_2, ind_2 = index_from_ellipse(u, du, r_x, r_y, 0)
        
        # d2u vs du
        r_x = intensity*em_du
        r_y = intensity*em_d2u
        xp_3, yp_3, ind_3 = index_from_ellipse(du, d2u, r_x, r_y, 0)

        ind = ind_1 | ind_2 | ind_3

        if verbose:
            
            print('Total to go = {}'.format(sum(ind)))

            plt.figure(figsize=(5, 15))

            ax = plt.subplot2grid((3, 20), (0, 0), rowspan=1, colspan=29)
            plt.plot(u, d2u, '.')
            plt.plot(u[ind_1], d2u[ind_1], '.')
            plt.plot(xp_1, yp_1)
            ax.grid(linestyle='-')

            ax = plt.subplot2grid((3, 20), (1, 0), rowspan=1, colspan=29)
            plt.plot(u, du, '.')
            plt.plot(u[ind_2], du[ind_2], '.')
            plt.plot(xp_2, yp_2)
            ax.grid(linestyle='-')
            
            ax = plt.subplot2grid((3, 20), (2, 0), rowspan=1, colspan=29)
            plt.plot(du, d2u, '.')
            plt.plot(du[ind_3], d2u[ind_3], '.')
            plt.plot(xp_3, yp_3)
            ax.grid(linestyle='-')
            
        return ind

    def calculate_ellipse(origin_x, origin_y, r_x, r_y, phi=0):
        """

        """
        
        ang = np.linspace(-np.pi, np.pi, 1000)

        x = origin_x + r_x*np.cos(ang)*np.cos(phi) - r_y*np.sin(ang)*np.sin(phi)
        y = origin_y + r_y*np.sin(ang)*np.cos(phi) + r_x*np.cos(ang)*np.sin(phi)

        x = np.append(x, x[0])
        y = np.append(y, y[0])
        
        return x, y 

    """
    Main function start
    """

    if np.any(np.isnan(u)):
        raise Exception('Fill nans before despiking')

    ind = phase_space_thres(u, fs, intensity=intensity, verbose=verbose)

    U = u.copy()
    if verbose:
        sn = np.sum(np.sum(np.isnan(U)))
        print('Nans in the data before despike: {}'.format(sn)) 
    
    U[ind] = np.nan
    di = np.isnan(U)
    if verbose:
        sn = np.sum(np.sum(di))
        print('Nans in the data after despike: {}'.format(sn)) 
    
    U = fillgaps(U, extrapFlg=True)
    if verbose:
        sn = np.sum(np.sum(np.isnan(U)))
        print('Nans in the data after the gapfill: {}'.format(sn)) 

    if verbose:
    
        plt.show()

        plt.figure(figsize=(15, 5))
        plt.plot(u)
        plt.plot(U)

    return U, di

def logical_clean(x, index):
    """
    Interpolate over values indicated by some index. 
    """

    xo = x.copy()
    xo[index] = np.nan

    xo = fillgaps(xo)

    return xo

def nan_helper(y):
    """Helper to handle indices and logical indices of NaNs.

    Input:
        - y, 1d numpy array with possible NaNs
    Output:
        - nans, logical indices of NaNs
        - index, a function, with signature indices= index(logical_indices),
          to convert logical indices of NaNs to 'equivalent' indices
    Example:
        >>> # linear interpolation of NaNs
        >>> nans, x= nan_helper(y)
        >>> y[nans]= np.interp(x(nans), x(~nans), y[~nans])
    """

    return np.isnan(y), lambda z: z.nonzero()[0]

def fillnans(y, first_gone=None):
    """
    Fill nans in a 1D or 2D np array.
    """

    if not type(None) == type(first_gone):
        y[first_gone::, :] = np.nan
    
    s = y.shape
    if len(s) == 1:
        y = y[:, None]
        n = s
        mm = 1
    elif len(s) == 2:
        nn, mm = s
    else:
        raise(Exception)
        
    for m in np.arange(0, mm):
                
        nans, x= nan_helper(y[:, m])

        if sum(nans)>0 and sum(nans)<len(nans):
            y[nans, m]= np.interp(x(nans), x(~nans), y[~nans, m])

    return y


def fillgaps(a, maxgap=np.inf, dim=0, extrapFlg=False):
    """
    Linearly fill NaN value in an array.

    Parameters
    ----------
    a : |np.ndarray|
      The array to be filled.

    maxgap : |np.ndarray| (optional: inf)
      The maximum gap to fill.

    dim : int (optional: 0)
      The dimension to operate along.

    extrapFlg : bool (optional: False)
      Whether to extrapolate if NaNs are found at the ends of the
      array.

    See Also
    =====

    interpgaps : Linearly interpolates in time.

    Notes
    =====

    This function interpolates assuming spacing/timestep between
    successive points is constant. If the spacing is not constant, use
    interpgaps.

    This was duplicated from the DOLFyn function of the same name. See lkilcher/DOLFyN on github. 

    """

    # If this is a multi-dimensional array, operate along axis dim.
    if a.ndim > 1:
        raise Exception("This functionality has been removed")

    a = np.asarray(a)
    nd = a.ndim
    if dim < 0:
        dim += nd
    if (dim >= nd):
        raise ValueError("dim must be less than a.ndim; dim=%d, rank=%d."
                         % (dim, nd))
    ind = [0] * (nd - 1)
    i = np.zeros(nd, 'O')
    indlist = list(range(nd))
    indlist.remove(dim)
    i[dim] = slice(None, None)
    # outshape = np.asarray(a.shape).take(indlist)
    # Ntot = np.product(outshape)
    i.put(indlist, ind)
    # k = 0

    gd = np.nonzero(~np.isnan(a))[0]

    # Here we extrapolate the ends, if necessary:
    if extrapFlg and gd.__len__() > 0:
        if gd[0] != 0 and gd[0] <= maxgap:
            a[:gd[0]] = a[gd[0]]
        if gd[-1] != a.__len__() and (a.__len__() - (gd[-1] + 1)) <= maxgap:
            a[gd[-1]:] = a[gd[-1]]

    # Here is the main loop
    if gd.__len__() > 1:
        inds = np.nonzero((1 < np.diff(gd)) & (np.diff(gd) <= maxgap + 1))[0]
        for i2 in range(0, inds.__len__()):
            ii = list(range(gd[inds[i2]] + 1, gd[inds[i2] + 1]))
            a[ii] = (np.diff(a[gd[[inds[i2], inds[i2] + 1]]]) *
                     (np.arange(0, ii.__len__()) + 1) /
                     (ii.__len__() + 1) + a[gd[inds[i2]]]).astype(a.dtype)

    return a

        
#%%

# ho = 1

#%%


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar  5 13:33:59 2019

@author: andrewzulberti
"""

import numpy as np 
import matplotlib.pyplot as plt
import scipy.optimize
import utm

import gpxpy
import gpxpy.gpx

def triangulate(x, 
                y, 
                ranges, 
                total_depth,
                responder_height,
                x_drop=None,
                y_drop=None, 
                name='Unnamed object', 
                coord_sys='latlon'):

    def get_circle(xc, yc, r):
    
        xs = np.arange(xc-(r-tol), xc+(r-tol), 0.01)
        ysq = np.power(r, 2) - np.power((xs-xc), 2)
        ys = np.sqrt(ysq) + yc
        ys_ = -np.sqrt(ysq) + yc

        ys = np.concatenate((ys, ys_, (ys[0],)), 0)
        xs = np.concatenate((xs, np.flip(xs, 0), (xs[0],)), 0)

        return xs, ys
    
    def calculate_distance(checkparams, verbose = False):
        
        xcheck = checkparams[0]
        ycheck = checkparams[1]

        d1 = np.min(np.sqrt((y1 - ycheck)**2 ++ (x1 - xcheck)**2))
        d2 = np.min(np.sqrt((y2 - ycheck)**2 ++ (x2 - xcheck)**2))
        d3 = np.min(np.sqrt((y3 - ycheck)**2 ++ (x3 - xcheck)**2))
        d = np.sqrt(d1**2 + d2**2 + d3**2)

        if verbose:
            print('Distance 1 is {0}'.format(d1))
            print('Distance 2 is {0}'.format(d2))
            print('Distance 3 is {0}'.format(d3))

            print('Total distance is {0}'.format(d))

        return d

    tol = 0.1 
    lex = 700 # Width of the final plot

    vert_disp = - responder_height

    assert(len(x)==len(y))
    assert(len(x)==len(ranges))
    assert(len(x)>=3)
    
    got_drop=False
    if not y_drop is None:
        if not x_drop is None:
            got_drop=True
            
    if coord_sys == 'latlon':
        if got_drop:
            x_drop, y_drop, zone1, zone2 = utm.from_latlon(y_drop, x_drop)

        for ii, (xi, yi) in enumerate(zip(x, y)):#np.arange(0, len(x_extra)):
            x[ii], y[ii], zone1, zone2 = utm.from_latlon(y[ii], x[ii])

    print('UTM Zone: {0:.0f}, {1:s}'.format(zone1, zone2)) 
    # print(zone1)
    # print(zone2)

    # Only triangulates ont the first 3...
    r1 = np.sqrt(np.power(ranges[0], 2) - np.power(vert_disp, 2))
    r2 = np.sqrt(np.power(ranges[1], 2) - np.power(vert_disp, 2))
    r3 = np.sqrt(np.power(ranges[2], 2) - np.power(vert_disp, 2))

    # ... others are 'extra' and are plotted but not used
    ranges_extra = ranges[3::]
    x_extra = x[3::]
    y_extra = y[3::]
    
    x1, y1 = get_circle(x[0], y[0], r1)
    x2, y2 = get_circle(x[1], y[1], r2)
    x3, y3 = get_circle(x[2], y[2], r3)

    fig, axes = plt.subplots(nrows=1, figsize=(10, 10))

    axes.set_xlabel('Easting (m)')
    axes.set_ylabel('Northing (m)')
    
    r3h, = plt.plot(x3, y3, 'g', label='Third Mark')
    r2h, = plt.plot(x2, y2, 'r', label='Second Mark')
    r1h, = plt.plot(x1, y1, 'b', label='First Mark')

    if got_drop:
        axes.set_xlim((x_drop-lex, x_drop+lex))
        axes.set_ylim((y_drop-lex, y_drop+lex))
        ph, = plt.plot(x_drop, y_drop, 'kx', label='Drop Marker')
    
    for ii in np.arange(0, len(y_extra)):

        re = np.sqrt(np.power(ranges_extra[ii], 2) - np.power(vert_disp, 2))

        xe, ye = get_circle(x_extra[ii], y_extra[ii], re)
        plt.plot(xe, ye, 'k:', label='Extra Marks')

#     calculate_distance([x_drop, y_drop])

    fh = lambda pos: calculate_distance(pos)

    if got_drop:
        print('Initial guess taken as drop location.')
        x0 = [x_drop, y_drop]
    else:
        print('Initial guess taken as average of all marks, regardless of range.')
        x0 = [np.mean(x), np.mean(y)]
        print(x0)
        
    locale = scipy.optimize.fmin(func=fh, x0=x0)
    # lat, long = utm.to_latlon(locale[0], locale[1], 50, 'K')
    lat, long = utm.to_latlon(locale[0], locale[1], zone1, zone2)

    foundh, = plt.plot(locale[0], locale[1], 'gx', label='Solution')

#     if got_drop:
#         d = calculate_distance(locale, verbose = True)
#         textt = name + ': {0:.0f}, {1:.0f} [error: {2:.0f} m]'.format(locale[0], locale[1], d)
#         textt = name + ': {0:.5f}, {1:.5f} [error: {2:.0f} m]'.format(lat, long, d)
#     else:
#         textt = name + ': {0:.0f}, {1:.0f}'.format(locale[0], locale[1])
#         textt = name + ': {0:.5f}, {1:.5f}'.format(lat, long)
    d = calculate_distance(locale, verbose = True)
    textt = name + ': {0:.0f}, {1:.0f} [Uncertainty: {2:.0f} m]'.format(locale[0], locale[1], d)
    textt = name + ': {0:.5f}, {1:.5f} [Uncertainty: {2:.0f} m]'.format(lat, long, d)

    print(textt)
    plt.title(textt)

    plt.gca().set_aspect('equal', adjustable='box') 
    axes.grid(color='k', linestyle=':', linewidth=1, zorder=-1)

#     axes.legend([ph, r1h, r2h, r3h, foundh], ['Drop Marker', 'Range 1', 'Range 2', 'Range 3', 'Solution'])
#     axes.legend()
    handles, labels = axes.get_legend_handles_labels()
    by_label = dict(zip(labels, handles))
    axes.legend(by_label.values(), by_label.keys(), loc='center left', bbox_to_anchor=(1, 0.5))

    fullpathsave = '{name} Triangulation.png'.format(name=name)
    fig.savefig(fullpathsave)  
    
    return [lat, long], locale


class mygpx():
    """
    Parser/handler for GPX ship tracks which has seen very little testing. 
    """

    # Shouldn't import like this, but going for it!
    

    def __init__(self, gpxfgile):

        self.gpxfgile = gpxfgile
        self.read()
        self.process()
        
    def read(self):
        
        f = open(self.gpxfgile, 'r')
        print('File open')

        gpx = gpxpy.parse(f)
        print('File parsed')
        
        self.gpx = gpx
        
    def process(self):

        gpx_x = []
        gpx_y = []
        gpx_z = []
        gpx_t = []

        for track in self.gpx.tracks:
            for segment in track.segments:
                
                x = np.array([point.longitude for point in segment.points])
                y = np.array([point.latitude for point in segment.points])
                z = np.array([point.elevation for point in segment.points])
                t = np.array([point.time for point in segment.points])

                gpx_x += [x]
                gpx_y += [y]
                gpx_z += [z]
                gpx_t += [t]
                
        
        self.gpx_x = gpx_x
        self.gpx_y = gpx_y
        self.gpx_z = gpx_z
        self.gpx_t = gpx_t
        
        #         for point in segment.points:
        #             print('Point at ({0},{1}) -> {2}'.format(point.latitude, point.longitude, point.elevation))

        # for waypoint in gpx.waypoints:
        #     print('waypoint {0} -> ({1},{2})'.format(waypoint.name, waypoint.latitude, waypoint.longitude))

        # for route in gpx.routes:
        #     print('Route:')
        #     for point in route.points:
        #         print('Point at ({0},{1}) -> {2}'.format(point.latitude, point.longitude, point.elevation))

# parse_solander_gpx(gpxfgile)
    def plot_gpx(self, plot_dict={}):

        for x, y in zip(self.gpx_x, self.gpx_y):
            plt.plot(x, y, linewidth=0.5, alpha=0.5)

        plt.gca().set_aspect('equal')
    


"""
Various tools for working with bottom landers. Ultimately looking to have frame interference and :

Functions
=========
    - XXX:   XXX
    - XXX:   XXX

"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import path as mplpath
import afloat 

strict = False # only way to get this to work with autoreload. 

def rectify_xy(x, y, theta, rotate):
    """
    Rotate arbitrary amount in x and y
    """
    
    if rotate:
        h = (np.pi/180)*-theta
    else:
        h = 0

    x_ =  x*np.cos(h) - y*np.sin(h)
    y_ =  x*np.sin(h) + y*np.cos(h)

    return x_, y_
       
class frame_obs:
    """
    Class for handling single instrument record. At the moment handles Nortek Signature and Nortek Vector only. 
    """

    name = 'Unnamed obs'
        
    x_offset = 0
    y_offset = 0
    z_offset = 0
    
    t = None
    u, v, w = None, None, None
    b1, b2, b3 = None, None, None
    
    x_axis_expected = 0 # X axis as expected when laying out frame. Must be noted in field notes. 
    
    heading_oem = 0 # Measured Heading in manufacturer coordinates
    pitch_oem   = 0 # Measured Pitch in manufacturer coordinates
    roll_oem    = 0 # Measured Roll in manufacturer coordinates
    
    known_instrs = ["signature", "sig", "vector", "vec"]
    _instr = 'signature'

    good_arcs = []
    bad_arcs = []
    arc_grow = 1.25

    title_loc='north'

    def __init__(self, name, x=0, y=0, z=0, x_axis_expected=0, heading_oem=0, u=None, v=None, w=None, **vargs):
        
        self.__dict__.update(**vargs)
        self.name = name
        
        self.x_offset = x
        self.y_offset = y
        self.z_offset = z
        
        self.x_axis_expected = x_axis_expected
        
        self.heading_oem = heading_oem
        
        self.upright = True
    
    def print_insrument_conventions(self):

        known = self.known_instrs
        instr = self.instr
        
        if instr.lower() in known:
            pass
        else:
            raise(Exception("Instrument mus be in {}".format(known)))

        print(instr+':')
        if instr.lower() in ["signature", "sig"]:
            print("  Heading is defined as the geogaphical direction of the 'x' axis")
            print("  The 'x' axis is defined as the geogaphical direction of BEAM 1")
            print("  Instrument is oriented 'up' if transducers are pointing up. ")

        if instr.lower() in ["vector", "vec"]:
            print("  Heading is defined as the geogaphical direction of the 'x' axis")
            print("  The 'x' axis is defined as the geogaphical direction of BEAM 1")
            print("  Instrument is oriented 'up' if the canister is above the transducers (transducers are pointing down). ")

    @property
    def instr(self):        
        
        instr = self._instr
        if instr.lower() in ["signature", "sig"]:
            instr = 'signature'

        if instr.lower() in ["vector", "vec"]:
            instr = 'vector'
        
        return instr
    
    # Define setter
    @instr.setter
    def instr(self, value):
        self._instr = value.lower()
        
    @property
    def beam_data(self):
        
        instr = self.instr
        
        if instr in ['vector']:
            beam_rots = [2*np.pi/3, 4*np.pi/3, 0]
            beam_names_if_up = ['B2', 'B3', 'B1', ]
            beam_names_if_down = ['B3', 'B2', 'B1', ]
        elif instr in ['signature']:
            # Just plot slanted beams
            beam_rots = [2*np.pi/4, 4*np.pi/4, 6*np.pi/4, 0]
            beam_names_if_up = ['B2', 'B3', 'B4', 'B1', ]
            beam_names_if_down = ['B3', 'B2', 'B4', 'B1', ]
        
        beam_names = beam_names_if_up
        if not self.upright:
            beam_names = beam_names_if_down
            
        return beam_rots, beam_names
    
        
    @property
    def axis_data(self):
        
        instr = self.instr
        
        if instr in ['vector']:
            rots_if_up = np.array([0, 1*np.pi/2])
            names = ['X', 'Y']
        elif instr in ['signature']:
            # Just plot slanted beams
            rots_if_up = np.array([0, 6*np.pi/4])
            names = ['X', 'Y']
      
        rots = rots_if_up
        if not self.upright:
            rots = -rots_if_up
            
        return rots, names
    
    def rotate(self, x, y, z):
        
        o = self
        p_oem, r_oem, h_oem, h_exp = o.pitch_oem, o.roll_oem, o.heading_oem, o.x_axis_expected
        
        h_oem = (np.pi/180)*(90-h_oem)
        h_exp = (np.pi/180)*(90-h_exp)
        
        p_oem = (np.pi/180)*(p_oem)
        r_oem = (np.pi/180)*(r_oem)
        
#         % Think these should be negative
        hh = -h_oem
        pp = -p_oem
        rr = -r_oem

        H = np.array([[np.cos(hh), np.sin(hh), 0],[ -np.sin(hh), np.cos(hh), 0],[0, 0, 1]])

        # % Make tilt matrix
        P = np.array([[np.cos(pp), -np.sin(pp)*np.sin(rr), -np.cos(rr)*np.sin(pp)],
             [0, np.cos(rr), -np.sin(rr)],  
             [np.sin(pp), np.sin(rr)*np.cos(pp),  np.cos(pp)*np.cos(rr)]])

        # % Make resulting transformation matrix
        xyz2enu = np.matmul(H, P)

        xyz = np.stack([x, y, z])

        enu = xyz2enu @ xyz

        x = enu[0, :]
        y = enu[1, :]
        z = enu[2, :]
        
        return x, y, z

    def plot(self, heading_length=200, frame_rotation=0):

        rotate_frame = True
        max_y = 0
        max_x = 0
        
        o = self
        x, y, z = o.x_offset, o.y_offset, o.z_offset
        p_oem, r_oem, h_oem, h_exp = o.pitch_oem, o.roll_oem, o.heading_oem, o.x_axis_expected
        
        if rotate_frame:
            h_exp += frame_rotation
    
        instr = o.instr
        upright = o.upright

        h_oem = (np.pi/180)*(90-h_oem)
        h_exp = (np.pi/180)*(90-h_exp)
        
        p_oem = (np.pi/180)*(p_oem)
        r_oem = (np.pi/180)*(r_oem)

        x, y = rectify_xy(x, y, frame_rotation, rotate_frame) 
        
        plt.plot(x, y, 'kx')

        # Plot expected heading
        hexp_x = heading_length*np.cos(h_exp)
        hexp_y = heading_length*np.sin(h_exp)
        
        xv = np.array([0, hexp_x])
        yv = np.array([0, hexp_y])
        
        plt.plot(x+xv, y+yv, 'b--', label='Instr. X expected')

        max_x = np.max([max_x, max(xv)])
        max_y = np.max([max_y, max(yv)])
        
        # Plot a circle
        t = np.linspace(0, 2*np.pi, 100);
        xc = np.cos(t)*heading_length/2
        yc = np.sin(t)*heading_length/2
        zc = t*0
        
        xc, yc, zc = self.rotate(xc, yc, zc)

        plt.plot(x+xc, y+yc, 'k--')
        plt.plot(x+xc[0], y+yc[0], 'k*')

        # Plot out all beams, ending with the 'Heading' beam
        beam_rots, beam_names = o.beam_data

        for br, bn in zip(beam_rots, beam_names):
            
#             hoem_x = heading_length*np.cos(h_oem+br)
#             hoem_y = heading_length*np.sin(h_oem+br)
            hoem_x = heading_length*np.cos(br)
            hoem_y = heading_length*np.sin(br)
        
            xv = np.array([0, hoem_x])
            yv = np.array([0, hoem_y])
            
            max_x = np.max([max_x, max(xv)])
            max_y = np.max([max_y, max(yv)])
            
            zv = yv*0
            
            xv, yv, zv = self.rotate(xv, yv, zv)
            
            plt.plot(x+xv, y+yv, '0.5', linestyle=':')

            plt.text(x+xv[-1], y+yv[-1], '  '+bn, fontsize='small')
            
        axis_rots, axis_names = o.axis_data

        for i, (ar, an) in enumerate(zip(axis_rots, axis_names)):
            
            if i==0:
                col = 'r'
            elif i==1:
                col = 'g'
            else:
                continue
                
            if an.lower()=='x':
                lab = 'Instr. X observed'
            elif an.lower()=='y':
                lab = 'Instr. Y observed'
            else:
                lab = None
                
            hoem_x = heading_length*np.cos(ar)
            hoem_y = heading_length*np.sin(ar)
        
            xv = np.array([0, hoem_x])
            yv = np.array([0, hoem_y])
            
            max_x = np.max([max_x, max(xv)])
            max_y = np.max([max_y, max(yv)])
            
            zv = yv*0
            
            xv, yv, zv = self.rotate(xv, yv, zv)
            
            plt.plot(x+xv, y+yv, col, linestyle='-', label=lab)

            plt.text(x+xv[-1], y+yv[-1], an+'  ', color=col, fontsize='small', ha='right', va='bottom')
            
#         plt.plot(x+xv, y+yv, 'r', label='OEM heading')
    
        def plot_arc(arc, col, lab):
    
            arcs = arc[0]
            arce = arc[1]
            
            assert(arcs<=arce)
                
            t = np.linspace(arcs, arce, 100);
            t = 90-t
            t *= np.pi/180
            grow = self.arc_grow 
            xc = np.hstack([np.cos(t)*heading_length*(grow)/2, np.cos(t[::-1])*heading_length/(grow*2)])
            yc = np.hstack([np.sin(t)*heading_length*(grow)/2, np.sin(t[::-1])*heading_length/(grow*2)])
            
            plt.fill(x+xc, y+yc, col, alpha=0.4, label=lab)
            
        
        for arc in self.good_arcs:

            plot_arc(arc, 'g', 'Frame interference: no')
            
        for arc in self.bad_arcs:

            plot_arc(arc, 'r', 'Frame interference: yes')
            
            plt.grid()
        
        plt.gca().set_aspect('equal')
        plt.grid('on')
        
        handles, labels = plt.gca().get_legend_handles_labels()
        by_label = dict(zip(labels, handles))
        plt.legend(by_label.values(), by_label.keys(), loc='center left', bbox_to_anchor=(1, 0.5))
        
        if self.title_loc.lower()=='north':
            plt.text(x, 1.1*max_y+y, '' + o.name + '\n[z=' + str(z) + ' mm]\n', ha='center', va='bottom', fontweight='bold')
        elif self.title_loc.lower()=='east':
            plt.text(1.1*max_x+x, y, '' + o.name + '\n[z=' + str(z) + ' mm]\n', ha='left', va='center', fontweight='bold')
        elif self.title_loc.lower()=='northeast':
            plt.text(max_x+x, max_y+y, '' + o.name + '\n[z=' + str(z) + ' mm]\n', ha='left', va='bottom', fontweight='bold')
        elif self.title_loc.lower()=='south':
            plt.text(x, -1.1*max_y+y, '' + o.name + '\n[z=' + str(z) + ' mm]\n', ha='center', va='top', fontweight='bold')
        elif self.title_loc.lower()=='west':
            plt.text(-1.1*max_x+x, y, '' + o.name + '\n[z=' + str(z) + ' mm]\n', ha='right', va='center', fontweight='bold')
        else:
            raise(Exception('Title location not recognised'))
        

    
class frame:
    obs = []
    
    frame_land_heading = 0
    
    # 2D Geometry of the frame. Default UWA Lander
    x = np.array([0, 1200, 2400])
    y = np.array([0, 2078, 0])
    
    a = np.sin(30*np.pi/180)*600/np.sin(60*np.pi/180)
    x_sub = [x/2+600]
    y_sub = [y/2+a]
    
    def init(self, x, y):
        obs = []
        self.x, self.y = x, y 
        
    def __validate__(self):
        for o in self.obs:
            print(type(o))

            if strict:
                assert(type(o)==frame_obs | type(o)==afloat.bottom_lander.frame_obs) # This fails with module autoreload
            else:
                pass
        
        print("This frame has {} observations".format(len(self.obs)))
    def __repr__(self):
        return self.__str__()
    
    def __str__(self):
        """
        Custom print method
        """
        
        self.__validate__()
        
        return 'Not doing anything'
    
    def add_obs(self, o):
        
        if strict:
            assert(type(o)==frame_obs | type(o)==afloat.bottom_lander.frame_obs) # This fails with module autoreload
        else:
            pass

        self.obs.append(o)
        pass
    
    def draw_old(self, rotate_frame=True, figsize=(7, 7)):
        
        heading_length = 200
        
        plt.figure(figsize=figsize)
        
        plt.plot(0, 0, 'r+')
        
        max_y = 0
        for o in self.obs:
            
            x, y, z, h_oem, h_exp = o.x_offset, o.y_offset, o.z_offset, o.heading_oem, o.x_axis_expected
            instr = o.instr
            upright = o.upright
            
            if rotate_frame:
#                 h_oem += self.frame_land_heading
                h_exp += self.frame_land_heading

            h_oem = (np.pi/180)*(90-h_oem)
            h_exp = (np.pi/180)*(90-h_exp)
            
            x, y = rectify_xy(x, y, self.frame_land_heading, rotate_frame) 
            
            plt.plot(x, y, 'kx')
            plt.text(x, y, '   ' + o.name + ' [z=' + str(z) + ' mm]')
            
            # Plot expected heading
            hexp_x = heading_length*np.cos(h_exp)
            hexp_y = heading_length*np.sin(h_exp)
            xv = np.array([x, x+hexp_x])
            yv = np.array([y, y+hexp_y])
#             xv, yv = self.rectify_xy(xv, yv, rotate_frame)  

            plt.plot(xv, yv, 'b--', label='Expected heading')
            
            # Plot out all beams, ending with the 'Heading' beam
            beam_rots, beam_names_if_up = o.beam_data
            
            for br, bm in zip(beam_rots, beam_names_if_up):
                hoem_x = heading_length*np.cos(h_oem+br)
                hoem_y = heading_length*np.sin(h_oem+br)
                xv = np.array([x, x+hoem_x])
                yv = np.array([y, y+hoem_y])
                plt.plot(xv, yv, '0.5', linestyle=':')
                
                plt.text(xv[-1], yv[-1], bm)
                
            plt.plot(xv, yv, 'r', label='OEM heading')
                       
        # Now plot the frame
        x, y = self.x, self.y
        x, y = rectify_xy(x, y, self.frame_land_heading, rotate_frame) 
        plt.fill(x, y, 'k', alpha=0.05)
        
        for x, y in zip(self.x_sub, self.y_sub):
            x, y = rectify_xy(x, y, self.frame_land_heading ,rotate_frame) 
            
            plt.fill(x, y, 'k', alpha=0.05, ec='k', ls='-')
        
        plt.grid('on')
        plt.gca().set_aspect('equal')
        plt.xlabel('X offset (mm)')
        plt.ylabel('Y offset (mm)')
        
        handles, labels = plt.gca().get_legend_handles_labels()
        by_label = dict(zip(labels, handles))
        # plt.legend(by_label.values(), by_label.keys(), loc='center left', bbox_to_anchor=(1, 0.5))
        plt.legend(by_label.values(), by_label.keys())
#         plt.legend()

        if rotate_frame:
            plt.title('Frame rotated as landed on seabed')
        else:
            plt.title('Assuming frame landed with nominal North')

        # plt.show()
        
        return plt.gcf()

    def draw(self, rotate_frame=True, figsize=(7, 7), heading_length=200):
        
        plt.figure(figsize=figsize)
        
        if rotate_frame:
            frame_rotation = self.frame_land_heading
        else:
            frame_rotation = 0
            
        plt.plot(0, 0, 'r+')
        
        for o in self.obs:
            
#             x, y = rectify_xy(x, y, rotate_frame) 
            o.plot(frame_rotation=frame_rotation, heading_length=heading_length)
            
        # Now plot the frame
        x, y = self.x, self.y
        x, y = rectify_xy(x, y, self.frame_land_heading, rotate_frame) 
        plt.fill(x, y, 'k', alpha=0.05)
        
        for x, y in zip(self.x_sub, self.y_sub):
            x, y = rectify_xy(x, y, self.frame_land_heading, rotate_frame) 
            
            plt.fill(x, y, 'k', alpha=0.05, ec='k', ls='-')
        
        plt.grid('on')
        plt.gca().set_aspect('equal')
        plt.xlabel('X offset (mm)')
        plt.ylabel('Y offset (mm)')
        
        handles, labels = plt.gca().get_legend_handles_labels()
        by_label = dict(zip(labels, handles))
        # plt.legend(by_label.values(), by_label.keys(), loc='center left', bbox_to_anchor=(1, 0.5))
        plt.legend(by_label.values(), by_label.keys())

#         plt.legend()

        if rotate_frame:
            plt.title('Frame rotated as landed on seabed')
        else:
            plt.title('Assuming frame landed with nominal North')

        return plt.gcf()    


#%%
 
# if True:
#     from ..utils import validation
# else:
#     import validation

import datetime
import numpy as np

import pandas as pd

# Matlab datum is 0-Jan-1990. Datetime won't cop that.
matlab_datum  = datetime.datetime.strptime('1-1-1900', '%d-%m-%Y') - datetime.timedelta(days=1)
my_datum = datetime.datetime(1990,1,1)

def linspacetime(start, end, n=None, dt_s=None):
    """
    Create a linearly spaced np array of datetime objects. Currently uses pandas but could be done 
    with linspace and one of the ordinal time tools if I ever start to care about dependencies.
    
    Inputs:
     - start: datetime object for start of array. 
     - end: datetime object for end of array. 
     - n: number of points in array [use None if using dt_s]
     - dt_s: number of seconds between points in the array [use None if using n]

     If specifying dt_s, array will start from start, and be spaced perfectly by dt_s, ending at or before
     end.
    """

    if n==None and dt_s==None:
        raise Exception('Must specify n or dt_s')
    elif (not n==None) and (not dt_s==None):
        raise Exception('Must specify only one of n or dt_s')
    elif not dt_s==None: # must calculate new end and n
        dt_hopeful = end-start 
        n_spaces = np.floor(dt_hopeful.seconds/dt_s)

        dt = datetime.timedelta(seconds=n_spaces*dt_s) # new total dt
        end = start + dt

        n = np.floor(dt.seconds/dt_s)+1
        
        # print('New end is: {}'.format(end))
        # print('New n is: {}'.format(n))
        # print('New t is: {}'.format(t))
        
    datelist = pd.date_range(start, end, periods=n).to_pydatetime()
    # datelist = pd.date_range(start, end, periods=n).to_list()

    return np.array(datelist)

def is_well_spaced(date, fs, epsilon=0.01, verbose=False):
    """
    A function to determine whether a time array is well spaced at fs. This is necessary to determine 
    if we wish to block 'blindly' average over time T = n/fs where n is a constant set numbers of 
    samples and fs is the sample frequency.
    
    Inputs:
     - date: numpy row vector of datetime objects. 
     - fs: sample frequncy in Hz
     
    Optional inputs:
     - epsilon: maximum deviation from the specified fs
    """

    # validation.is_np_date_rowvector(date)

    if type(date[0]) == datetime.datetime:
        max_dt = np.max(np.diff(date))
        min_dt = np.min(np.diff(date))
    elif type(date[0]) == np.datetime64:
        # max_dt = datetime.timedelta(seconds=np.max(np.diff(date)).item().total_seconds())
        # min_dt = datetime.timedelta(seconds=np.min(np.diff(date)).item().total_seconds())
        max_dt = datetime.timedelta(seconds=np.max(np.diff(date))/np.timedelta64(1, 's'))
        min_dt = datetime.timedelta(seconds=np.min(np.diff(date))/np.timedelta64(1, 's'))

    else:
        print(type(date[0]))
        raise Exception

    max_allow = datetime.timedelta(seconds=1/fs)*(1+epsilon)
    min_allow = datetime.timedelta(seconds=1/fs)*(1-epsilon)

    if max_dt > max_allow:
        raise Exception('Maximum time spacing of {} greater than allowable [{}].'.format(max_dt, max_allow))
    elif min_dt < min_allow:
        raise Exception('Minimum time spacing of {} smaller than allowable [{}].'.format(min_dt, min_allow))
    else:
        if verbose:
            print('Time vector is well spaced')

def datetime_2_timestamp(date_in):

    date_out = np.array([datetime.datetime.timestamp(date_in[i]) for i in np.arange(0, len(date_in))])

    return date_out

def timestamp_2_datetime(date_in):

    date_out = np.array([datetime.datetime.fromtimestamp(date_in[i]) for i in np.arange(0, len(date_in))])
    date_out = np.array(date_out)

    return date_out

def datetime64_2_timestamp(date_in):

    date_out = datetime64_2_datetime(date_in)
    date_out = datetime_2_timestamp(date_out)

    return date_out
    
def timestamp_2_datetime64(date_in):

    date_out = timestamp_2_datetime(date_in)
    date_out = datetime_2_datetime64(date_out)

    return date_out
    
def datetime64_2_datetime(date_in):
    
    date_out = [datetime64_2_datetime_single(date_in[i]) for i in np.arange(0, len(date_in))]
    date_out = np.array(date_out)

    return date_out

def datetime64_2_datetime_single(date_in):
    
    timestamp = ((date_in - np.datetime64('1970-01-01T00:00:00')) / np.timedelta64(1, 's'))
    date_out = datetime.datetime.utcfromtimestamp(timestamp)
    
    return date_out

def datetime_2_datetime64(date_in):
    
    date_out = [np.datetime64(date_in[i]) for i in np.arange(0, len(date_in))]
    date_out = np.array(date_out)

    return date_out

def datetime_2_datetime64_single(date_in):
    
    date_out = np.datetime64(date_in)

    return date_out

def str_to_dt(string, format_lang = 'python', dateformat = 'dd-mmm-yyyy HH:MM:SS:FFF'):

    if format_lang.lower() == 'matlab':
        dateformat = matformat_2_py(dateformat)

    return [datetime.datetime.strptime(string[i], dateformat) for i in np.arange(0, len(string))]
    
def num2date_lk(mpltime):
    """
    Reimplementation of datetime num2date to stop date errors. That occur continually when updating dolfyn or datetime. 
    
    This was copied from the Levi Kilcher's Dolfyn package, hence the lk suffix. 
    """
    
    if isinstance(mpltime, np.ndarray):
        try:
            n = len(mpltime)
        except TypeError:
            pass
        else:
            out = np.empty(n, dtype='O')
            for idx, val in enumerate(mpltime.flat):
                out[idx] = num2date_lk(val)
            out.shape = mpltime.shape
            return out
        
    if np.isnan(mpltime):
        return None
    return datetime.datetime.fromordinal(int(mpltime)) + datetime.timedelta(days=mpltime % 1)


def matplotlibtime_2_matlabdatenum():

    pass

def matlabdatenum_2_matplotlibtime():

    pass

def datenum_2_matlabdatenum():

    pass

def matlabdatenum_2_datenum():

    pass

def matformat_2_py(matformat):
    """
    Simple hack code to allow me to use matlab date formats pn python. At the moment I 
    just use common formats but should ultimately switch to a regext type approach to 
    find and replace. Not going for that as I should just learn the python formats. 
    """

    if matformat == 'dd-mmm-yyyy HH:MM:SS':
        pyformat = '%d-%b-%Y %H:%M:%S'
    elif matformat == 'dd-mmm-yyyy HH:MM:SS:FFF':
        pyformat = '%d-%b-%Y %H:%M:%S:%f'
    else:
        raise FormatException('Have not hard coded this date format yet')

    print('For personal reference, {0} translates to {1}'.format(matformat, pyformat))

    return pyformat

def FormatException(Exception):

    pass

def tests():

    n = 100
    ans = linspacetime(pd.datetime.today(), pd.datetime.today()+datetime.timedelta(1), n=100)
    if not len(ans) == n:
        raise(Exception)

    n = 100
    dt = datetime.timedelta(seconds=10.001)
    dt_s=1/10
    ans = linspacetime(pd.datetime.today(), pd.datetime.today()+dt, dt_s=dt_s)

    is_well_spaced(ans, 1/dt_s)

    ans = datetime_2_datetime64(ans)

    is_well_spaced(ans, 1/dt_s)

    print('There are {} elements in this {} second period spaced at {}.'.format(len(ans), dt.seconds, dt_s))

    print('Time Module OK')

pass

def np_datetime64_strftime(t, fmt):
    """
    Format a np.datetime64 or list of np.datetime64 objects as a string.
    
    returns: 
        list of strings
        
    """
    
    if not type(t) == list:
        t = [t]
        
    d = [pd.to_datetime(str(t_)).strftime(fmt) for t_ in t]
    
    return d


def datetime64_mean(timein):
    """
    Calculate mean of a sequence of datetime/datetime64 return as datetime64
    """

    return datetime64_from_seconds_since([np.mean(seconds_since(timein))])

def datetime_mean(timein):
    """
    Calculate mean of a sequence of datetime/datetime64 return as datetime
    """

    return datetime_from_seconds_since([np.mean(seconds_since(timein))])

def datetime_from_seconds_since(tsec, basetime=my_datum, low_prec=True):
    """
    Converts a list or array of seconds since "basetime" into an array of datetime object
    From MRayson SFODA
    """

    if low_prec:
        return np.array([basetime + datetime.timedelta(seconds=int(i)) for i in tsec])
    else:
        return np.array([basetime + datetime.timedelta(microseconds=int(i*1000)) for i in tsec])

def datetime64_from_seconds_since(tsec, basetime=my_datum, low_prec=True):
    """
    Converts a list or array of seconds since "basetime" into an array of np.datetime64 object
    From MRayson SFODA
    """

    basetime = datetime_2_datetime64_single(basetime)
    
    if low_prec:
        return np.array([basetime + np.timedelta64(int(i), 's') for i in tsec])
    else:
        return np.array([basetime + np.timedelta64(int(i*1e9), 'ns') for i in tsec])
        
def strftime(timein, fmt):
    """
    Just a 'shortcut' to datetime.datetime.strftime that operates on both datetime.datetime and np.datetime64
    """
    if isinstance(timein, datetime.datetime):
        isdatetime = True
    elif isinstance(timein, np.datetime64):
        isdatetime64 = True
        timein = datetime64_2_datetime_single(timein)
    else:
        error
        rrrr
        
    return timein.strftime(fmt)
        
def seconds_since(timein, basetime=my_datum):
    """
    Converts a list or array of datetime object into an array of seconds since "basetime"
    
    From MRayson SFODA
    """
    
    # Convert the time to seconds

    # Workout the input type
    isarray = False
    isdatetime = False
    isdatetime64 = False
    if isinstance(timein, list) or isinstance(timein, np.ndarray):
        isarray = True
        if isinstance(timein[0], datetime.datetime):
             isdatetime = True
        elif isinstance(timein[0], np.datetime64):
             isdatetime64 = True

    # Could use the funcitons above but this seems fine for now. 
    if isarray and isdatetime64:
        time0 = np.datetime64(basetime).astype('<M8[us]')
        tsec = ((timein.astype('<M8[us]') - time0)).astype(np.float64)*1e-6

    elif isarray and isdatetime:
        tsec = np.array([(t-basetime).total_seconds() for t in timein])

    else: # Assume its a datetime.datetime object (this is not good..)
        if isinstance(timein, datetime.datetime):
            tsec = (timein-basetime).total_seconds()

        else:
            time0 = np.datetime64(basetime).astype('<M8[us]')
            try:
                #raise Exception, 'Time conversion error'
                # This case shouldn't arise
                tsec = ((timein.astype('<M8[us]') - time0)).astype(np.float64)*1e-6
            except:
                tsec = (timein-time0).total_seconds() # TimeDelta object

    return tsec

def time_interp1(time_out, time_in, data_in):
    """
    np.interp for timeseries data. 
    
    """
    
    # Convert times to an integer
    time_out_sec = seconds_since(time_out)
    time_in_sec = seconds_since(time_in)
    
    return np.interp(time_out_sec, time_in_sec, data_in)

#%%


# -*- coding: utf-8 -*-
"""
Created on Mon Jul 30 15:17:42 2018

@author: 20270917
"""
#%%

import matplotlib.pyplot as plt
import pandas
import numpy as np
import xarray as xr
from matplotlib.dates import num2date, date2num
import matplotlib
import datetime
import os 
import pdb 

from zutils import qc_conventions
from zutils import file as zfile

raise(Exception("Moved to AFLOAT"))

default_attrs = {
    'title': '', 
    'institution': 'The University of Western Australia', 
    'institution_division': 'Ocean Dynamics', 
    'source': '', 
    'project': '', 
    'history': '', 
    'references': '', 
    'comment': '', 
    'Conventions': 'CF-1.7', 
    'trip': '', 
    'trip_deployed': '', 
    'site': '', 
    'site_station': '', 
    'instrument_make': '',  
    'instrument_model': '',
    'instrument_serial_number': '',
    'raw_file_name': '',
    'raw_file_directory': '',
    'raw_file_attributes': '',                  # This is a string which holds the attributes of any raw file taken in by the. 
    'last_export_file_name': '',                # When a netcdf is loaded, this should be cleared
    'last_export_directory': '',                # When a netcdf is loaded, this should be cleared
    'last_load_file_name': '',                  # When a netcdf is loaded, this should be overwritten with the load_name
    'last_load_directory': '',                  # When a netcdf is loaded, this should be overwritten with the load_name
    'outfile_append': '', 
    'disclaimer': '',
    'nominal_latitude': '',
    'nominal_longitude': '',
    'nominal_site_depth': '',
    'pressure_sensor_height_asb': '',
    'nominal_instrument_height_asb': '',
    'nominal_instrument_orientation': '',
    'timezone': '',
    'process_level': '',
    'is_profile_data': 0} 

def parse_infile(infile, verbose=True):
    """
    Take a string ( folder/file ), or 2 element list  ( [folder, file] ), and return the folder and file.
    """

    if type(infile)==str:
        if verbose:
            print('Infile is a string')
        folder, file = os.path.split(infile)
    elif type(infile) in [list, tuple]:
        if not len(infile) == 2:
            raise(Exception('The infile must be a string or a length 2 sequence'))
        else:
            folder, file = infile
    else:
        raise(Exception('The infile must be a string or a length 2 sequence'))
        
    return folder, file
    
def _from_netcdf(infile, classhandler):
    """
    Pass straight to the main xrwrap load method.

    Inputs:
        - infile
        - classhandler - the xrwrap class to open the file
    """
   
    folder, file = parse_infile(infile)

    ds = xr.open_dataset(os.path.join(folder, file))

    ds.attrs['last_load_file_name']      = file
    ds.attrs['last_load_directory']      = folder

    # print(ds)
    
    rr = classhandler(ds)

    return rr, ds

class xrwrap():
    
    ######
    # SHOULD MINIMISE THE NUMBER OF PROPERTIES HERE, BECAUSE THESE DON'T EXPORT TO NETCDF OR LOAD FROM NETCDF 
    ######
    # folder = ''
    # file_ = ''

    # Certain irreversible operations will lead to modification of the output file name to prevent overwrite. 
    auto_file_append = '' 
    
    so = None
    eo = None

    first_good = None
    last_good = None

    _qc_conv = qc_conventions.base()
    _default_attrs = default_attrs
    _required_attrs = default_attrs.keys()
    _outname = None
    
    # attrs = default_attrs

    default_user = 'UWA'

    verbose = False

    @property
    def _obj(self):
        """
        Little trick that may or may not work for xarray accessors.
        """
        return self.ds

    @property
    def fullpath(self, i=0):
        
        raise(Exception('Don''t use this fullpath property anymore. Switching to a more detailed system involving dataset attributes. '))

        if type(self.file_) == list:
            return '{folder}/{file_}'.format(folder=self.folder, file_=self.file_[i])
        else:
            return '{folder}/{file_}'.format(folder=self.folder, file_=self.file_)

    @property
    def attrs(self):
        return self.ds.attrs

    def generate_outname(self, keep_name='False', fileappend=''):
        """
        Function to autogenerate a file name. Should not do this continuously.
        """
        use_profiler_covention = False
        if self.is_profile_data or use_profiler_covention:
            convention = ['project',
                            'trip',
                            'site_station',
                            'instrument_model',
                            'instrument_serial_number',
                            'raw_file_name',
                            'outfile_append']
        else:

            convention = ['project',
                            'trip',
                            'site_station',
                            'nominal_instrument_height_asb',
                            'instrument_model',
                            'instrument_serial_number',
                            'outfile_append']

        outname = naming_conv(self.attrs, convention=convention)
        return outname

    # @property
    def outname(self, fileappend=''):
        """
        generate an outname only if outname is empty.
        """
        if self._outname is None:
            self._outname = self.generate_outname(fileappend='')

        return self._outname

    def load(self, infile):
        """
        Initialise from netcdf rather than from a raw input file. 
        """

        raise(Exception("Don't load like this anymore. Load from outside the actual object."))

        folder, file = parse_infile(infile)

        nc_file = os.path.join(folder, file)
        
        ds = xr.open_dataset(nc_file)

        ds.attrs['last_load_file_name']      = file
        ds.attrs['last_load_directory']      = folder

        self.store_raw_file_attributes(ds)
        self.ds = ds

        # self.attrs = ds.attrs
        # self.ds.attrs = {}

        print('Loaded from NC. Class attributes taken from NC. NC attrs cleared.')

        pass

    def wrap(self, ds):
        """
        Initialise from an xarray dataset. Got to be carful here that the dataset is of the exact format or lots of errors will be thrown. Could used this if 2 NC files
        are merged outside of here and then brought back in. That's the only usage I can immagine.  
        """

        self.ds = ds
        if False: # Still not sure the best point to do this.
            self.store_raw_file_attributes(ds)

        print('Wrapped an existing xarray dataset. Class attributes taken from the dataset. Dataset attrs cleared.')
        
    def export(self, naming_method=None, export_directory=None, final=False):
        """
        Base export class. Overloading should really be avoided.

        Inputs:
            - naming_method: specifies how the output name is to be generated
                * 'raw_file'   = use the name and directory of the raw file, but with extension *.nc
                * 'last_load'  = use the name and directory of the last netcdf loaded
                * 'convention' = use the inbuilt naming convention for the filename. Directory will be last load if that is not blank, else it will be the raw_dir.
                * None         = use 'last_load' if it exists, else use 'raw_file'
            - export_directory: specify new directory if you don't want the program to choose
        """

        outdir, outname = self.get_export_location(naming_method=naming_method, export_directory=export_directory)

        self.attrs['last_export_directory'] = outdir
        self.attrs['last_export_file_name'] = outname

        ## 
        if final:
            self.ds.attrs['Disclaimer'] = self.disclaimer
            # outname = outname + 'finalised'

        self.ds.close() # Force close

        nc_file = '{folder}//{file}'.format(folder=outdir, file=outname)

        print('Exporting {}'.format(nc_file))
        
        self.ds.to_netcdf(path=nc_file)

        return self.ds # It may be useful in many cases to return somethng here. Just return self.sd for consistency with subclasses.  

    @property
    def best_export_dir(self):
        """
        Directory will be last load if that is not blank, else it will be the raw_dir.
        """

        outdir  = self.attrs['last_load_directory']
        if outdir == '':
            print('Best export dir is the raw_file_directory.')
            outdir  = self.attrs['raw_file_directory']
        else:
            print('Best export dir is the last_load_directory.')

        
        return outdir

    def get_export_location(self, naming_method=None, export_directory=None):
        """
        Base export class. Overloading should really be avoided.

        Inputs:
            - naming_method: specifies how the output name is to be generated
                * 'raw_file'   = use the name and directory of the raw file, but with extension *.nc
                * 'last_load'  = use the name and directory of the last netcdf loaded
                * 'convention' = use the inbuilt naming convention for the filename. Directory will be last load if that is not blank, else it will be the raw_dir.
                * None         = use 'last_load' if it exists, else use 'raw_file'
            - export_directory: specify new directory if you don't want the program to choose
        """

        if naming_method is None:
            if not len(self.attrs['last_load_directory']) == 0:
                naming_method = 'last_load'
            else:
                naming_method = 'raw_file'

        if naming_method.lower() == 'convention':
            print('Generating filename from naming convention.')
            outname = self.generate_outname()
            outname = zfile.drop_extension(outname) + '.nc'
            outdir  = self.best_export_dir

        elif naming_method.lower() == 'last_load':
            outdir  = self.attrs['last_load_directory']
            outname = self.attrs['last_load_file_name']
            print('Will save file using last load file name and directory.')

        elif naming_method.lower() == 'raw_file':
            outdir  = self.attrs['raw_file_directory']
            outname = self.attrs['raw_file_name']
            # Replace extension
            outname = zfile.drop_extension(outname) + '.nc'
            print('Will save file using raw file name and directory.')

        ## Use new directory if specified
        if not export_directory is None:
            outdir = export_directory
            print('Will to user input directory.')

        return outdir, outname

    @property
    def fullpath_last_export(self):
        """
        Full path that you last exported to.
        """

        return os.path.join(self.attrs['last_export_directory'], self.attrs['last_export_file_name'])

    def time_trim(self, first_good, last_good):

        if first_good is None:
            first_good = self.ds.time.values[0]

        if last_good is None:
            last_good = self.ds.time.values[-1]

        self.first_good = first_good
        self.last_good = last_good

        print('Good data from {} to {}'.format(first_good, last_good))

        # This sometimes fails if the DataSet time is timezone aware and the trim times
        # are timezone naive. Dolfyn seems to throw both:
            #  type(rr.ds.time.values[1]) = pandas._libs.tslibs.timestamps.Timestamp
            #  type(rr.ds.time.values[1]) = numpy.datetime64
        # and only the former is datetime aware. Ultimately I'll be going away from Dolfyn. 
        # there is no fix for this, user must be aware of this. 
        self.ds = self.ds.sel(time=slice(self.first_good, self.last_good))

        print('Trimmed Time')

    @property  
    def is_profile_data(self):
        
        if 'is_profile_data' in self.attrs:
            return self.attrs['is_profile_data']
        else:
            return False

    @property  
    def qc_conv(self):
        return self._qc_conv

    @qc_conv.setter
    def qc_conv(self, new_qc_conv):
        """
        Set function for _qc_conv
        """
        if isinstance(new_qc_conv, qc_conventions):
            self._qc_conv = new_qc_conv
        else:
            raise(Exception(("Must ba a QC Convention object")))

    def run_qc_compliance(self, flag_name):
        """
        Wrapper for the qc_conv.run_compliance method.

        Basically it checks that your QC flag [ds data_var with the name flag_name] complies with the QC conventions specified by the _qc_conv attribute.
        """

        self.qc_conv.run_compliance(self.ds[flag_name])

    def associate_qc_flag(self, var_name, flag_name):
        """
        Assign a QC flag to a DataArray, and add the flag to the dataset if necessary.
        Inputs:
                - var_name: The name of the data variable which will have a QC flag associated. Cannot QC dimensions [coords].
                - flag_name: The name of QC flag to be associated/added. The prefix 'qc_' will be added automatically
        """

        # Add the qc_ prefix
        flag_name = 'qc_' + flag_name
        
        # Check if var_name exists
        if not var_name in self.ds.data_vars:
            raise(Exception('Variable does not exist'))

        var_shape = self.ds[var_name].values.shape

        # Check if flag_name exists
        if not flag_name in self.ds.data_vars:
            # if not create the flag
            flag_array = self.ds[var_name].copy()
            flag_array.attrs={}
            flag_array.values[:] = -999
            self.ds[flag_name] = flag_array

            # flag = -999*np.ones_like(self.ds[var_name])
            # self.ds[flag_name] = flag
            self.qc_conv.run_compliance(self.ds[flag_name])
        elif not self.ds[flag_name].values.shape == var_shape:
            # If so check the size of the variable against the size of the flag
            raise(Exception("The selected QC flag does not have the right dimensions for this variable."))

        # Check this variable doesn't already have a qc code. Multiples not yet allowed.
        if 'qc_variable' in self.ds[var_name].attrs:
            raise(Exception("A QC variable has already been associated to the variable {}!".format(var_name)))

        for i in self.ds.data_vars:
            if 'is_qc_flag' in self.ds[i].attrs and 'associated data variables' in self.ds[i].attrs: # It might be a QC flag
                if  self.ds[i].attrs['is_qc_flag']: # It's a QC flag
                    if var_name in self.ds[i].attrs['associated_data_variables'].split(';'):
                        raise(Exception("The variable {} has already been associated to the QC flag {}!".format(var_name, i)))

        # Associate this flag to this variable
        self.ds[flag_name].attrs['associated_data_variables'] += ';'+var_name
        self.ds[var_name].attrs['qc_variable'] = flag_name

        # Trim leading ; if necessary
        if self.ds[flag_name].attrs['associated_data_variables'][0] == ';':
            self.ds[flag_name].attrs['associated_data_variables'] = self.ds[flag_name].attrs['associated_data_variables'][1::]
    
    @property  
    def valid_qc_flags(self):
        """
        All valid QC flags. Checks for:
            - 'qc_' prefix
            - presence of 'is_qc_flag' attribute
            - value of 'is_qc_flag' attribute

        The checks could certainly be more thorough.
        """

        out = []
        for i in self.ds.data_vars:
            if len(i) < 4:
                continue
            if not i[0:3] == 'qc_':
                continue
            if not 'is_qc_flag' in self.ds[i].attrs:
                continue
            if self.ds[i].attrs['is_qc_flag']:
                out.append(i)

        return(out)

    def set_pressure_sensor_height(self, ps_hasb):

        self.update_attribute('pressure_sensor_height_asb', ps_hasb)

    def get_pressure_sensor_height(self):

        pressure_sensor_height_asb = self._attrs['pressure_sensor_height_asb']

        if type(pressure_sensor_height_asb) == str:
            pressure_sensor_height_asb = np.nan
            
        return pressure_sensor_height_asb

    @property
    def pressure_sensor_height(self):

        return self.get_pressure_sensor_height()

    def advance_time_mins(self, advance_mins=None, dataset_name='ds', time_name='time', comment=""):
        """
        Function to shift clock and log the change. 

        Inputs:
            advance_mins is the number of minutes to advance the clock [i.e. move to a later time].
            dataset_name is the name of the dataset this change will be applied to. Default is "ds".
            time_name is the name of the time variable this change will be applied to. Default is "time".
             
        """

        if advance_mins == 0 or advance_mins is None or np.isnan(advance_mins):
            print('Not advancing time')
            return
            
        string = 'Advanced the time variable "{}" by {} minutes with user comment "{}"'.format(time_name, advance_mins, comment)
        self.add_comment(self.default_user, string, ds_name='ds', data_var=None)
        
        print(string)

        dataset = getattr(self, dataset_name)
        dataset = dataset.assign_coords({time_name: dataset[time_name] + np.timedelta64(advance_mins,'m')})
        setattr(self, dataset_name, dataset)

    def offset_variable(self, offset_in_var_units, variable_name, dataset_name='ds',  comment=""):
        """
        Modify raw data by adding a set offset. 

        The offset is sepcified in whatever units the variable is specified in.  

        """
    
        if offset_in_var_units == 0 or offset_in_var_units is None or np.isnan(offset_in_var_units):
            print('Not offsetting variable')
            return
        
        raise(Exception("Not implemented"))
        
    def overwrite_variable(self, new_val_in_var_units, variable_name, dataset_name='ds',  comment=""):
        """
        Modify variable by overwriting its value(s) with a constant or equal sized array.  
        
        The new value(s) is sepcified in whatever units the variable is specified in. 
        """
        
        string = 'Setting the variable "{}" to {} [same units] with user comment "{}"'.format(variable_name, 
                                                                                              new_val_in_var_units, comment)
        self.add_comment(self.default_user, string, ds_name='ds', data_var=None)
        
        print(string)

        dataset = getattr(self, dataset_name)
        dataset[variable_name].values[:] = new_val_in_var_units
        

    def update_qc_flag_dict(self, flag_name, index_dict, flag_value, comment=None, delete_raw=False, verbose=False):
        """
        This is a base function to update a QAQC flag. Inputs:
            - flag_name: The name of the netcdf variable corrsponding to the QC flag being edited. Can be:
                            - a single flag name
                            - a list of flag names
                            - '*'
                        if '*' is used all QC flags will be updated 

            - index_name: The name of the netcdf variable which is being used to as an index to 
                            identify which points to in the QC flag variable are to be edited.
            - start: The first point in the index to change the value of
            - end: The last point in the index to change the value of
            - flag_value: new value which the QC flag is to become wherever the index is between start and end 
        """
        
        if delete_raw:
            error

        if isinstance(flag_name, str): 
            if flag_name == '*': 
                flag_names = self.valid_qc_flags
            else:
                flag_names = [flag_name]
        else: # Assume list
            flag_names = flag_name

        for flag_name in flag_names:

            flag_dims = [dim for dim in self.ds[flag_name].dims]
            index_check = [index_name in self.ds[flag_name].dims for index_name in index_dict.keys()]
            if not np.all(index_check):
                continue

            logind = np.ones_like(self.ds[flag_name].values.astype(bool))
            for index_name in index_dict.keys():

                index = expand_dims(self.ds[index_name], flag_dims)

                start, end = index_dict[index_name]

                logind1 = index >= start
                logind = np.logical_and(logind, logind1)

                logind2 = index <= end
                logind = np.logical_and(logind, logind2)

            # Note that the logical index is reversed here.
            self.ds[flag_name] = self.ds[flag_name].where(~logind, flag_value)
                
            ind = np.where(logind)[0]
            logind = self.ds[flag_name] > 0

            if not comment is None: # Log comment on the QAQC.
                string = 'Flagged {} values with code "{}" and user comment "{}"'.format(len(self.ds[index_name].values)-len(ind), flag_value, comment)

                self.add_comment(self.default_user, string, data_var=flag_name)

            if self.verbose:
                
                print(ind)
                print(len(ind))
                
                print('There are {} flagged data points.'.format(np.sum(logind.values)))
                print('')        

    def update_qc_flag(self, flag_name, index_name, start, end, flag_value, comment=None, delete_raw=False, verbose=False, allow_back_compat=True):
        """
        This is a base function to update a QAQC flag. Inputs:
            - flag_name: The name of the netcdf variable corrsponding to the QC flag being edited. Can be:
                            - a single flag name
                            - a list of flag names
                            - '*'
                        if '*' is used all QC flags will be updated 

            - index_name: The name of the netcdf variable which is being used to as an index to 
                            identify which points to in the QC flag variable are to be edited.
            - start: The first point in the index to change the value of
            - end: The last point in the index to change the value of
            - flag_value: new value which the QC flag is to become wherever the index is between start and end 
        """
        
        if not allow_back_compat:
            raise(Exception('The function update_qc_flag is no longer available. Use the function update_qc_flag_dict or try again with the allow_back_compat option. '))
            
        index_dict = {}
        index_dict[index_name] = [start, end]

        self.update_qc_flag_dict(flag_name, index_dict, flag_value, comment=comment, delete_raw=delete_raw, verbose=verbose)

    def update_qc_flag_logical(self, flag_name, index_name, logical_index, flag_value, comment=None, delete_raw=False, verbose=False):
        """
        This is a base function to update a QAQC flag. Inputs:
            - flag_name: The name of the netcdf variable corrsponding to the QC flag being edited. Can be:
                            - a single flag name
                            - a list of flag names
                            - '*'
                        if '*' is used all QC flags will be updated 

            - index_name: The name of the netcdf variable which is being used to as an index to 
                            identify which points to in the QC flag variable are to be edited.
            - logical_index: Logical index for points to change the value of
            - flag_value: new value which the QC flag is to become wherever the index is between start and end 
        """

        if delete_raw:
            error

        if isinstance(flag_name, str): 
            if flag_name == '*': 
                flag_names = self.valid_qc_flags
            else:
                flag_names = [flag_name]
        else: # Assume list
            flag_names = flag_name

        for flag_name in flag_names:

            if not index_name in self.ds[flag_name].coords:
                continue

            # Note that the logical index is reversed here.
            self.ds[flag_name] = self.ds[flag_name].where(~logical_index, flag_value)

            ind = np.where(logical_index)[0]
            logind = self.ds[flag_name] > 0

            if self.verbose:
                print()

            if not comment is None: # Log comment on the QAQC.
                # string = 'Flagged {} values with code "{}" and user comment "{}"'.format(len(self.ds[index_name].values)-len(ind), flag_value, comment)
                string = 'Flagged {} values with code "{}" and user comment "{}"'.format(len(ind), flag_value, comment)
                    
                self.add_comment(self.default_user, string, data_var=flag_name)

            if self.verbose:
                print(ind)
                print(len(ind))

                print('There are {} flagged data points.'.format(np.sum(logind.values)))
                print('')   

    def flip_qc_value(self, flag_name, value_out=-999, value_in=0, verbose=False):
        """
        Flip one QC value for another. By default it will flip -999 to 0

        flag_name can be:
                                - a single flag name
                                - a list of flag names
                                - '*'
                            if '*' is used all QC flags will be updated 
        
        NOTE: doesn't actually check that the input variable is QC convention compliant. 
        
        """
        
        if isinstance(flag_name, str): 
                if flag_name == '*': 
                    flag_names = self.valid_qc_flags
                else:
                    flag_names = [flag_name]
        else: # Assume list
            flag_names = flag_name

        for flag_name in flag_names:

            logind = self.ds[flag_name] == value_out
            logind = ~logind

            self.ds[flag_name] = self.ds[flag_name].where(logind, value_in)

            if self.verbose:
                self.get_fig_text()

                logind = self.ds[flag_name] > 0
                print('There are {} positive flag points.'.format(np.sum(logind.values)))
                logind = self.ds[flag_name] < 0
                print('There are {} negative flag points.'.format(np.sum(logind.values)))
                print('')   

    def get_qaqc_var(self, var_name):
        """
        Retrun a QAQC'd copy of the data array by var_name. This just routes to the main function.
        """
        
        da = get_qaqc_var(self.ds, var_name)
        
        return da 

    def has_dates(self):

        if type(self.so) == type(None):
            return False
        elif type(self.eo) == type(None):
            return False
        else:
            return True

    def get_fig_text(self, fileappend=''):
        """
        Get a text string to describing the dataset.
        """

        delim = ' | '
        s = ''
        k = self.attrs.keys()
        
        if 'project' in k:
            s = s + delim + self.attrs['project']
            
        if 'trip_recovered' in k:
            s = s + delim + self.attrs['trip_recovered']
            
        if 'site' in k:
            s = s + delim + self.attrs['site']
            
        if not fileappend == '':
            s = s + delim + fileappend
            
        s = s[len(delim)::]

        print(s)

        return s

    def add_history(self, author, string, ds_name=None, data_var=None):
        """
        Add CF Compliant string to the history attribute of a Dataset or DataArray.

        Use the data_var kwarg to specify a DataArray. Omit or set to None to work with the whole Dataset.
        """

        attr = 'history'
        self.add_string(attr, author, string, ds_name=ds_name, data_var=data_var)
            
    def add_comment(self, author, string, ds_name='ds', data_var=None):
        """
        Add CF Compliant string to the comments attribute of a Dataset or DataArray.

        Use the data_var kwarg to specify a DataArray. Omit or set to None to work with the whole Dataset.
        """

        attr = 'comment'
        self.add_string(attr, author, string, ds_name=ds_name, data_var=data_var)
            
    def add_string(self, attr, author, string, ds_name='ds', data_var=None):
        """
        Add CF Compliant string to an attribute of a Dataset or DataArray.

        Use the data_var kwarg to specify a DataArray. Omit or set to None to work with the whole Dataset.
        """
        
        if data_var is None:
            obj = getattr(self, ds_name)
        else:
            obj = getattr(self, ds_name)[data_var]

        new_string = datetime.datetime.now().isoformat() + ': ' + '[{}]'.format(author) + ' ' + string
        
        if not attr in obj.attrs.keys():
            obj.attrs[attr] = new_string
        elif obj.attrs[attr] in ['', '?', 'blank']:
            obj.attrs[attr] = new_string
        else:
            obj.attrs[attr] = obj.attrs[attr] + ';' + new_string
            
        return new_string

    def update_attribute(self, attribute_name, attribute_value, ds_name='ds', data_var=None, strict=True):
        """
        This function updates the hidden attributes property of the class. The attribute must exist in the default attributes dictionary.
        """

        if data_var is None:
            obj = getattr(self, ds_name)
        else:
            obj = getattr(self, ds_name)[data_var]

        if (attribute_name in self._default_attrs) or (not strict):
            if self.verbose:
                print('Setting attribute "{}" to "{}"'.format(attribute_name, attribute_value))
            obj.attrs[attribute_name] = attribute_value
        else:
            raise(Exception('{} is not a valid attribute.'.format(attribute_name)))

        # Repeated calls are made here, shouldn't be too slow. 
        self.pIMOS_assign_coords()

    def update_attributes_with_dict(self, attribute_dict):
        """
        This function updates the hidden attributes property of the class. The attribute must exist in the default attributes dictionary.
        """

        for attribute_name in attribute_dict.keys():
            
            self.update_attribute(attribute_name, attribute_dict[attribute_name])

    
    def pIMOS_assign_coords(self):
        """
        Assign some additional coords to the file. These coords use deployment knowledge, so are not available at first creation of the ds.  

        Note: at the moment this doesn't alter the coords on any data variables.

        Function can be called repeatedly, it simply overwrites any coords.

        """
            
        try:
            lat_nom = self.ds.attrs['nominal_latitude']
            self.ds = self.ds.assign_coords({'lat_nom': lat_nom,})
        except:
            # print('Failed assigning nominal_latitude')
            pass
            
        try:
            lat_nom = self.ds.attrs['nominal_longitude']
            self.ds = self.ds.assign_coords({'lon_nom': lat_nom,})
        except:
            # print('Failed assigning nominal_longitude')
            pass

        try:
            z_nom = self.ds.attrs['nominal_site_depth'] + self.ds.attrs['nominal_instrument_height_asb'] 
            self.ds = self.ds.assign_coords({'z_nom': z_nom,})
        except:
            # print('Failed assigning z_nom')
            pass
            
    
    @property
    def rawattrs(self):
        """
        Return the raw file attribute string as a dict.
        
        This is sloppy, and will likely fail under many data types. 
        """
        
        array = np.array
        
        self.attrs['raw_file_attributes'] = "{'_type': 'ADCP', 'name': 'unrecognized firmware version', 'sourceprog': 'instrument', 'prog_ver': 51.4, 'config': '01000001-11001010', 'beam_angle': 20, 'numbeams': 4, 'beam_freq_khz': 300, 'beam_pattern': 'convex', 'orientation': 'up', 'simflag': 'real', 'n_beam': 4, 'n_cells': 46, 'pings_per_ensemble': 60, 'cell_size_m': 2.0, 'blank_m': 1.76, 'prof_mode': 1, 'corr_threshold': 0, 'prof_codereps': 5, 'min_pgood': 0, 'evel_threshold': 2000, 'sec_between_ping_groups': 1.0, 'coord': '00000000', 'coord_sys': 'beam', 'use_pitchroll': 'no', 'use_3beam': 'no', 'bin_mapping': 'no', 'xducer_misalign_deg': 0.0, 'magnetic_var_deg': 0.0, 'sensors_src': '01111101', 'sensors_avail': '00111101', 'bin1_dist_m': 4.2, 'xmit_pulse': 2.38, 'water_ref_cells': array([1, 5]), 'fls_target_threshold': 255, 'xmit_lag_m': 0.49}"
        
        return eval(self.attrs['raw_file_attributes'])
        
    def get_raw_file_attributes(self):
        """
        Returns the attributes from rawfile as a dict.
        """

        return eval(self.attrs['raw_file_attributes'])

    def store_raw_file_attributes(self, ds):
        """
        Function that will pull any valid attributes from the dataset into the wrapper, and will
        push any remaining into the 'raw_file_attributes' attribute.
        """
        
        print('STORING RAW FILE ATTRIBUTES')
        input_attributes = ds.attrs.copy()
        keys = input_attributes.keys()

        invalid_attributes = {} # These can't be added to the dataset 
        ds.attrs = self._default_attrs  
        
        for attribute in keys:
            
            if attribute in xrwrap._default_attrs:
                self.update_attribute(attribute, input_attributes[attribute])
            else:
                invalid_attribute = input_attributes[attribute]
                invalid_attributes[attribute] = invalid_attribute
                print(attribute.upper())
        pass

        if not len(invalid_attributes) == 0:

            if not ds.attrs['raw_file_attributes'] == '':
                raise(Exception('This operation can be performed only once'))

            # self.add_string('raw_file_attributes', self.default_user, str(invalid_attributes), ds_name='ds', data_var=None)
            self.attrs['raw_file_attributes'] = str(invalid_attributes)

    def enforce_these_attrs(self, enforced_attributes):
        """
        Enforces certain global attributes to have certain values. Error will be thrown if the attribute are not
        blank and different to what is specified:

        Inputs:
            - enforced_attributes (dict): dictionary of with attribute names as keys, and the enforced value as value.  
        """

        # CF Compliance
        if self.verbose:
            print(self.attrs)

        for attr in enforced_attributes.keys():
            if self.attrs[attr] == '':
                self.update_attribute(attr, enforced_attributes[attr])
            else:
                assert(self.attrs[attr]==enforced_attributes[attr])  

    def check_attrs(self):
        """
        Check that all attributes in the dataset are in the default attributes, and vice versa.
        
        TO DO:
            - There should probably be a list of attributes that need not be blank or a warning given. 
        """
        
        for key in self.ds.attrs.keys():
            assert(key in self._default_attrs)
        for key in self._default_attrs:
            assert(key in self.ds.attrs.keys())
    
    def parse_attributes(self, ds_to_check=None):
        """
        Generic function to check consistency of attributes between the object itself and the properties of the class.
        """

        raise(Exception('We no longer store attributes this way - this function should not be used'))

        if ds_to_check is None:
            ds_to_check = self.ds

        print("Parsing attributes.")
        for i in ds_to_check.attrs.keys():
            if i in self.attrs.keys():
                print("{} is both a property of the object and an attribute of the dataset".format(i))
                if ds_to_check.attrs[i] == self.attrs[i]:
                    print("     ... and they are equal")
                else:
                    print("     ... and they NOT are equal!!!")

        ds_to_check.attrs = self.attrs

    @property
    def disclaimer(self):
        """

        """

        disc = """These data were prepared for a specific purpose at a specific site, which may or may not be disclosed in the data or metadata. 
        Use of these data for any purpose, without the express written consent of the author is not permitted. 
        The data are provided WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
        The author is not liable in any way for consequences arising from any use of these data or any derivatives in any application including but not limited to design or decision-making processes.
        """

        return disc
  
def get_qaqc_var(ds, var_name):
    """
    Retrun a QAQC'd copy of the data array by var_name
    """
    
    da = ds[var_name].copy() # return a QCd copy

    if not 'qc_variable' in ds[var_name].attrs:
        print('Variable {} has no QAQC flag. Returning raw.'.format(var_name))
    else:    
        flag_name = ds[var_name].attrs['qc_variable'] # Currently assumes only one qc_variable
        ind = ds[flag_name].values > 0 # Currently does not remove suspect data. Should be an option.
        print('Blanking {} values.'.format(np.sum(ind)))

        values=da.values
        values[ind] = np.nan
        da.values=values
    
    return da             

### Generic XR stuff

def select_calendar_month(X, year_month, timename='time'):
    """
    Return a subset of the Dataset or DataArray limited to the specified year and month. Uses rather crude indexing methods. I know 
    there are much more sopistocated inbuilt functions but I'm not familiar with them.
    """

    def calendar_month(year, month):
        """
        For a given year and month return the date of the begining of the month and the date of the beginning of the next month
        """
        start = datetime.datetime(year, month, 1)
        if month == 12:
            end = datetime.datetime(year+1, 1, 1)
        else:
            end = datetime.datetime(year, month+1, 1)
        print(start)
        print(end)
        return start, end
        
    year, month = year_month
    
    start, end = calendar_month(year, month)
    
    # Gotta do better than this
    if timename.lower() == 'time':
        X_cm = X.sel(time = slice(start, end))
    elif timename.lower() == 'time_wave':
        X_cm = X.sel(time_wave = slice(start, end))
    # Gotta do better than this
    
    return X_cm, [start, end]

def split_by_timegap(X, timename='time', hours=1):
    """
    Split a DataArray along the time dimension based on a set time gap.
    
        Returns a list of DataArrays with the gaps removed.
        
    """
        
    time = X[timename].values
    dt = np.diff(time)

    # print(len(time))
        
    i = [int(ii) for ii in np.where(dt>np.timedelta64(hours, 'h'))[0]] + [len(time)-1]
    i = [-1] + list(set(i))
    i.sort()
    
    # print('Split index')
    # print(i)
    
    Xs = [X.isel({timename: np.arange(i[j]+1, i[j+1])}) for j in np.arange(0, len(i)-1)]
    
    return Xs

def expand_dims(array, full_dims):
    """
    Expands out an array to help with broadcasting.

    Inputs: 
        - array: Array you wish to expand
        - full_dims: List containing names of dimensions of the array you wish to expand out to

    Returns:
        - vals: values of the array expanded into the dimensions within full_dims

    """
    
    expand_list = []
    for flag_dim in full_dims:
        if flag_dim in array.coords:
            expand_list.append(np.arange(0, len(array[flag_dim])))
        else:
            expand_list.append(None)

    vals = array.values[expand_list]

    return vals

def get_dim_index(array, dim_name):
    """
    Find which axis of a DataArray corresponds to a certain dimension name.
    """
        
    coord_names =[x for x in array.coords ]
    dim_index =[i for i in np.arange(0, len(coord_names)) if coord_names[i]==dim_name]    
    
    if len(dim_index) == 1:
        dim_index=dim_index[0]
    else:
        dim_index = None
        
    return dim_index

def naming_conv(attrs, convention=None):
    """
    Create a 'conventional name' from a list of attributes defining the naming convention.      
    """
    
    if convention is None:
        convention = ['project',
                    'trip',
                    'site_station',
                    'instrument_model',
                    'instrument_serial_number']
    
    name = ''
    for attribute in convention:
        val = attrs[attribute]
        if type(attrs[attribute]) == str:
            # Check if it has been ; delimited, in which case return a special output
            split_attr = val.split(';')
            split_attr = set(split_attr) # Just the unique ones
            n = len(split_attr)
            if n == 1:
                pass
            else:
                val = 'multi({})_'.format(n)+attribute

        elif np.isnan(val):
            val = ''
        else:
            val = str(int(attrs[attribute]*100))
            
        if val == '':
            val = 'no_'+attribute

        if '.' in val:
            val = zfile.drop_extension(val)

        name += '[{}]_'.format(val)
    
    name = name[0:-1]
    
    return name


# %%



#!/usr/bin/env python

from setuptools import setup, find_packages
from setuptools.dist import Distribution

class BinaryDistribution(Distribution):
    def is_pure(self):
        return False

		
setup(name='afloat',
      version='0.0.3',
      description='A Fairly useful Library of Ocean Analysis Tools',
      author='Andrew Zulberti',
      author_email='andrew.zulberti@gmail.com',
      packages=find_packages(),
      install_requires=['numpy',
                        'matplotlib', 
                        'netcdf4', 
                        'scipy',
                        'xarray'],
      license='unlicensed to all but author',
      include_package_data=True,
      distclass=BinaryDistribution,
    )



#%%
 
# if True:
#     from ..utils import validation
# else:
#     import validation

import datetime
import numpy as np
import xarray as xr
import pandas as pd
 
# Alias for backward compat
from zutils.xrwrap import *

#%% Some code for adding seasons to xarray datasets

def add_season_custom(ds, month_season_dict, custom_season_name):
    """
    Add custom season to xarray dataset.
    
    Inputs:
        ds - xarray dataset. Time dimension must be called "time".
        month_season_dict - dictionary where the keys are the season names, and the values are the months in that season.
        
    """
    
    allmonths = []
    # Check all 12 months are covered with no overlap
    for season in month_season_dict:
        allmonths += month_season_dict[season]
        
    print(allmonths)
    allmonths = [int(i) for i in allmonths]
    l = len(allmonths)
    
    allmonths = np.unique(np.array(allmonths)).tolist()
    if not len(allmonths)==l:
        raise(Exception('Duplicate months.'))
    
    if len(allmonths)<12:
        raise(Exception('Not all months are covered.'))
    if not min(allmonths) == 1:
        raise(Exception('Invalid months'))
    if not max(allmonths) == 12:
        raise(Exception('Invalid months'))
    
    # Initialise
    custom_season = np.array(['none' for i in ds.time])
    month = ds['time.month']
        
    for season in month_season_dict:
        for m in month_season_dict[season]:
            custom_season[ds['time.month']==m] = season
        
    da = xr.DataArray(custom_season, dims={'time': ds.time})
    ds[custom_season_name] = da
        
def add_season_pilbara_wet_dry(ds):
    """
    Add a Pilbara wet/dry season variable to a dataset. No transition seasons. 
    
    Inputs:
        ds - xarray dataset. Time dimension must be called "time".
        
    """
        
    # Pilbara wet dry with no transitions
    month_season_dict = {'wet': [11, 12, 1, 2, 3, 4],
                         'dry': [5, 6, 7, 8, 9, 10]
    }

    custom_season_name = 'Pilbara wet dry'
    
    add_season_custom(ds, month_season_dict, custom_season_name)
    
def add_season_nw_monsoon(ds):
    """
    Add a northwest monsoon season variable to a dataset. 
    
    Inputs:
        ds - xarray dataset. Time dimension must be called "time".
        
    """
    
    # Pilbara wet dry with no transitions
    month_season_dict = {'NW': [9, 10, 11, 12, 1, 2],
                         't1': [3, 4],
                         'SE': [5, 6, 7],
                         't2': [8]
    }

    custom_season_name = 'NW Monsoon'
    
    add_season_custom(ds, month_season_dict, custom_season_name)
    
# %%
